#include "WBeeRSA.hpp"
#include "rsa.h"

static void WriteKey( FILE *keyFile, const WChar* variable, const WArray<uint32> &buffer );
static void KeyPair2Buffer( const rsakp *kp, WArray<uint32> &buffer );
static void Buffer2KeyPair( const int *buffer, rsakp *kp );
static void PublicKey2Buffer( const rsapk *pk, WArray<uint32> &buffer );
static void Buffer2PublicKey( const int *buffer, rsapk *pk );
static void Signature2String( mp32number *signature, WString &string );
static void String2Signature( const WString &string, mp32number *signature );
static WString PrepareMessage( const WString &signature );

WBool RSA_KeyGenerator( const WChar* keyFile, WInt size )
{
    randomGeneratorContext rngc;
    if (randomGeneratorContextInit(&rngc, randomGeneratorDefault()) == 0)
    {
        rsakp kp;
        rsakpInit(&kp);
        rsakpMake(&kp, &rngc, (size + 31) / 32 );
        
        //Save private key (rsakp) and public key (rsapk)
        WArray<uint32> buffer;
        FILE *keyFileHandle = fopen( keyFile, "w" );
        KeyPair2Buffer( &kp, buffer );
        WriteKey( keyFileHandle, "private_key", buffer );
        fputs( "\n", keyFileHandle );
        PublicKey2Buffer( (const rsapk *)&kp, buffer );
        WriteKey( keyFileHandle, "public_key", buffer );
        fclose( keyFileHandle );
        
        //Free allocated memory
        rsakpFree(&kp);
        randomGeneratorContextFree(&rngc);
        
        return TRUE;
    }
    return FALSE;
}

WString RSA_Sign( const WString &message, const int* privateKey )
{
    //Read Key
    rsakp kp;
    rsakpInit(&kp);
    Buffer2KeyPair( privateKey, &kp );

    //Create the Signature
    mp32number digest, s;
    mp32nzero( &digest );
    mp32nzero( &s );
    WString localMessage = PrepareMessage( message );
    mp32ninit(&digest, localMessage.GetLength() / sizeof(uint32), (const uint32*)localMessage.GetText() );

    rsapri(&kp, &digest, &s);
    WString signature;
    Signature2String( &s, signature );

    mp32nfree(&digest);
    mp32nfree(&s);
    rsakpFree(&kp);
    
    return signature;
}

WBool RSA_Verify( const WString &message, const WString &signature, const int* publicKey )
{
    //Read Key
    rsapk pk;
    rsapkInit(&pk);
    Buffer2PublicKey( publicKey, &pk );
    
    //Verify the signature
    mp32number digest, s;
    mp32nzero( &digest );
    mp32nzero( &s );
    WString localMessage = PrepareMessage( message );
    mp32ninit(&digest, localMessage.GetLength() / sizeof(uint32), (const uint32*)localMessage.GetText() );
    String2Signature( signature, &s );

    int rc = rsavrfy( &pk, &digest, &s);

    mp32nfree(&digest);
    mp32nfree(&s);
    rsapkFree(&pk);

    return rc != 0;
}


//Some usefull functions
static void write_mp32barrett( WArray<uint32> &, const mp32barrett *element );
static void write_mp32number( WArray<uint32> &, const mp32number *element );
static void read_mp32barrett( const int *buffer, int &index, mp32barrett *element );
static void read_mp32number( const int *buffer, int &index, mp32number *element );

static void WriteKey( FILE *keyFile, const WChar* variable, const WArray<uint32> &buffer )
{
    WString writeMessage;
    writeMessage.Sprintf("const int %s[%d] = \n{\n", variable, buffer.GetCount() );
    fputs( writeMessage, keyFile );
    writeMessage = "";
    for( int i=0; i < buffer.GetCount(); ++i )
    {
        WString help;
        help.Sprintf("0x%8.8lx", buffer[i] );
        writeMessage += help;
        if( i != buffer.GetCount() - 1 )
            writeMessage += ", ";
        if( i % 8 == 7 )
        {
            fputs( WString("    ") + writeMessage + "\n", keyFile );
            writeMessage = "";
        }
    }
    if( ! writeMessage.GetEmpty() )
    {
        fputs( WString("    ") + writeMessage + "\n", keyFile );
        writeMessage = "";
    }
    fputs( "};\n", keyFile );
}

static void KeyPair2Buffer( const rsakp *kp, WArray<uint32> &buffer )
{
    buffer.SetCount( 0 );
    
	write_mp32barrett( buffer, &kp->n );
	write_mp32number( buffer, &kp->e );
	write_mp32number( buffer, &kp->d );
	write_mp32barrett( buffer, &kp->p );
	write_mp32barrett( buffer, &kp->q );
	write_mp32number( buffer, &kp->d1 );
	write_mp32number( buffer, &kp->d2 );
	write_mp32number( buffer, &kp->c );
}

static void Buffer2KeyPair( const int *buffer, rsakp *kp )
{
    int index = 0;
    
	read_mp32barrett( buffer, index, &kp->n );
	read_mp32number( buffer, index, &kp->e );
	read_mp32number( buffer, index, &kp->d );
	read_mp32barrett( buffer, index, &kp->p );
	read_mp32barrett( buffer, index, &kp->q );
	read_mp32number( buffer, index, &kp->d1 );
	read_mp32number( buffer, index, &kp->d2 );
	read_mp32number( buffer, index, &kp->c );
}

static void PublicKey2Buffer( const rsapk *pk, WArray<uint32> &buffer )
{
    buffer.SetCount( 0 );

	write_mp32barrett( buffer, &pk->n );
	write_mp32number( buffer, &pk->e );
}

static void Buffer2PublicKey( const int *buffer, rsapk *pk )
{
    int index = 0;

	read_mp32barrett( buffer, index, &pk->n );
	read_mp32number( buffer, index, &pk->e );
}

static void Signature2String( mp32number *signature, WString &string )
{
    string = "";
    char *signatureString = (char *)signature->data;
    
    for( int i=0; i < (signature->size * sizeof(uint32)); ++i )
    {
        char zeichen[2] = "0";
        zeichen[0] = ( (signatureString[i] & 0xF0) >> 4 );
        if( zeichen[0] >= 0xA )
            zeichen[0] += 'A' - 0xA;
        else
            zeichen[0] += '0';
        string += zeichen;

        zeichen[0] = signatureString[i] & 0xF;
        if( zeichen[0] >= 0xA )
            zeichen[0] += 'A' - 0xA;
        else
            zeichen[0] += '0';
        string += zeichen;
    }
}

static void String2Signature( const WString &string, mp32number *signature )
{
    mp32nsize( signature, (string.GetLength() / 2) / sizeof(uint32));
    char *signatureString = (char *)signature->data;
    
    for( int i=0; i < string.GetLength() / 2; ++i )
    {
        char zeichen;
        zeichen = string[i * 2];
        if( (zeichen >= 'A') && (zeichen <= 'F' ))
            zeichen -= 'A' - 0xA;
        else
            zeichen -= '0';
        signatureString[i] = (zeichen << 4) ;

        zeichen = string[i * 2 + 1];
        if( (zeichen >= 'A') && (zeichen <= 'F' ))
            zeichen -= 'A' - 0xA;
        else
            zeichen -= '0';
        signatureString[i] = signatureString[i] | zeichen;
    }
}

static WString PrepareMessage( const WString &signature )
{
    WString l_signature(signature, TRUE);
    while( (l_signature.GetLength() % sizeof(uint32)) != 0 )
    {
        l_signature += " ";
    }
    return l_signature;
}


//Usefull for the usefull functions
static void write_mp32barrett( WArray<uint32> &buffer, const mp32barrett *element )
{
    int count = buffer.GetCount();
    buffer.SetCount( count + 1 + element->size + (element->size + 1) );
    
    buffer[count] = element->size;
    for( int i=0; i < element->size; ++i )
        buffer[count + 1 + i] = element->modl[i];
    for( i=0; i < (element->size + 1); ++i )
        buffer[count + 1 + element->size + i] = element->mu[i];
}

static void write_mp32number( WArray<uint32> &buffer, const mp32number *element )
{
    int count = buffer.GetCount();
    buffer.SetCount( count + 1 + element->size );
    
    buffer[count] = element->size;
    for( int i=0; i < element->size; ++i )
        buffer[count + 1 + i] = element->data[i];
}

static void read_mp32barrett( const int *buffer, int &index, mp32barrett *element )
{
    const WChar *valBuffer = (const WChar *)buffer;
    uint32 mp32barrettSize;
    memcpy( &mp32barrettSize, (void const *)&(valBuffer[index]), sizeof(element->size) );
    index += sizeof( element->size);
    
    mp32binit( element, mp32barrettSize );
    
    memcpy( element->modl, (void const *)&(valBuffer[index]), element->size * sizeof(uint32) );
    index += element->size * sizeof(uint32);
    
    memcpy( element->mu, (void const *)&(valBuffer[index]), (element->size + 1) * sizeof(uint32) );
    index += (element->size + 1) * sizeof(uint32);
}

static void read_mp32number( const int *buffer, int &index, mp32number *element )
{
    const WChar *valBuffer = (const WChar *)buffer;
    uint32 mp32numberSize;
    memcpy( &mp32numberSize, (void const *)&(valBuffer[index]), sizeof( element->size) );
    index += sizeof( element->size);
    
    mp32nsize(element, mp32numberSize);
    
    memcpy( element->data, (void const *)&(valBuffer[index]), element->size * sizeof(uint32) );
    index += element->size * sizeof(uint32);
}

